<!DOCTYPE html>
<html>
    <head>
        <meta charset="utf-8">
        <title>conda-store</title>
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <script defer src="static/conda-store-ui/main.js"></script>
        <link href="static/conda-store-ui/main.css" rel="stylesheet">
    </head>
    <body>
    </body>
</html>
