from conda_store_server.server.views.api import router_api  # noqa
from conda_store_server.server.views.registry import router_registry  # noqa
from conda_store_server.server.views.ui import router_ui  # noqa
from conda_store_server.server.views.metrics import router_metrics  # noqa
from conda_store_server.server.views.conda_store_ui import router_conda_store_ui  # noqa
